import argparse
import os

from cosmpy.aerial.client import LedgerClient
from cosmpy.aerial.faucet import FaucetApi
from cosmpy.aerial.wallet import LocalWallet
from cosmpy.crypto.keypairs import PrivateKey
from ptpython import embed
from jenesis.config import Config
from jenesis.contracts.detect import detect_contracts
from jenesis.contracts.monkey import MonkeyContract
from jenesis.contracts.observer import DeploymentUpdater
from jenesis.keyring import query_keychain_items, query_keychain_item
from jenesis.network import run_local_node


def load_config(args: argparse.Namespace) -> dict:
    project_path = os.getcwd()

    # check that we are actually running the command from the project root
    if not os.path.exists(os.path.join(project_path, "jenesis.toml")):
        # pylint: disable=all
        raise RuntimeError("Please run command from project root or create project first")

    cfg = Config.load(project_path)
    contracts = detect_contracts(project_path)

    shell_globals = {}
    contract_instances = {}

    if args.profile is not None and args.profile not in cfg.profiles:
        print(f'Invalid profile name. Expected one of {",".join(cfg.profiles.keys())}')
        return

    profile_name = args.profile or cfg.get_default_profile()

    selected_profile = cfg.profiles.get(profile_name)
    if selected_profile is not None:
        if selected_profile.network.is_local:
            run_local_node(selected_profile.network)

        # build the ledger client
        client = LedgerClient(selected_profile.network)

        print(f'Network: {selected_profile.network.name}')

        print('Detecting contracts...')

        for contract in contracts:
            print("C", contract)

            # skip contracts that we have not compiled
            if contract.digest() is None:
                continue

            # select the metadata for the contract
            selected_contract = selected_profile.deployments.get(contract.name)

            if selected_contract is not None:
                address = selected_contract.address
                code_id = selected_contract.code_id
            else:
                code_id = None
                address = None

            monkey = MonkeyContract(
                contract,
                client,
                address=address,
                code_id=code_id,
                observer=DeploymentUpdater(
                    cfg,
                    project_path,
                    profile_name,
                    contract.name,
                ),
                init_args=selected_contract.init,
            )

            contract_instances[contract.name] = monkey

        print('Detecting contracts...complete')
        
        shell_globals["ledger"] = LedgerClient(selected_profile.network)
        if selected_profile.network.faucet_url is not None:
            shell_globals["faucet"] = FaucetApi(selected_profile.network)

    wallets = {}
    for key in query_keychain_items():
        try:
            info = query_keychain_item(key)
            wallets[key] = LocalWallet(PrivateKey(info.private_key))
        except Exception:
            print(f"Failed to import local key '{key}'")

    shell_globals["cfg"] = cfg
    shell_globals["project_path"] = project_path
    shell_globals["profile"] = profile_name
    shell_globals["contracts"] = {contract.name: contract for contract in contracts}
    shell_globals["wallets"] = wallets
    for (name, instance) in contract_instances.items():
        shell_globals[name] = instance

    return shell_globals


def run(args: argparse.Namespace):
    shell_globals = load_config(args)
    shell_globals.update(globals())
    embed(shell_globals, vi_mode=False, history_filename=".shell_history")


def add_shell_command(parser):
    shell_cmd = parser.add_parser("shell")
    shell_cmd.add_argument(
        "-p", "--profile", default=None, help="The profile to use"
    )
    shell_cmd.set_defaults(handler=run)



from typing import Optional, Any
from abc import ABC, abstractmethod

import grpc
from cosmpy.aerial.client import LedgerClient
from cosmpy.aerial.contract import LedgerContract, _compute_digest
from cosmpy.aerial.tx_helpers import SubmittedTx
from cosmpy.aerial.wallet import Wallet
from cosmpy.crypto.address import Address
from cosmpy.protos.cosmwasm.wasm.v1.query_pb2 import QueryCodeRequest
from jsonschema import validate
from jenesis.contracts import Contract

INSTANTIATE_MSG = "instantiate_msg"
EXECUTE_MSG = "execute_msg"
QUERY_MSG = "query_msg"

class InstantiateArgsError(RuntimeError):
    pass


class ContractObserver(ABC):
    @abstractmethod
    def on_code_id_update(self, code_id: int):
        pass

    @abstractmethod
    def on_contract_address_update(self, address: Address):
        pass


class MonkeyContract(LedgerContract):
    def __init__(
        self,
        contract: Contract,
        client: LedgerClient,
        address: Optional[Address] = None,
        digest: Optional[bytes] = None,
        code_id: Optional[int] = None,
        observer: Optional[ContractObserver] = None,
        init_args: Optional[dict] = None,
    ):
        # pylint: disable=super-init-not-called
        self._contract = contract
        self._path = contract.binary_path
        self._client = client
        self._address = address
        self._observer = observer
        self._init_args = init_args

        # build the digest
        if contract.binary_path is not None:
            self._digest = _compute_digest(contract.binary_path)
        elif digest is not None:
            self._digest = digest
        else:
            raise RuntimeError('Unable to determine contract digest')

        # determine the code id
        if code_id is not None:
            self._code_id = self._find_contract_id_by_digest_with_hint(code_id)
        else:
            self._code_id = self._find_contract_id_by_digest(self._digest)

    def store(
            self,
            sender: Wallet,
            gas_limit: Optional[int] = None,
            memo: Optional[str] = None,
    ) -> int:
        code_id = super().store(sender, gas_limit=gas_limit, memo=memo)

        # trigger the observer if necessary
        if self._observer is not None:
            self._observer.on_code_id_update(code_id)

        return code_id

    def instantiate(
            self,
            code_id: int,
            args: Any,
            sender: Wallet,
            label: Optional[str] = None,
            gas_limit: Optional[int] = None,
            admin_address: Optional[Address] = None,
            funds: Optional[str] = None,
    ) -> Address:
        # if no args provided, insert init args from configuration
        if args is None:
            if self._init_args is not None:
                args = self._init_args
            else:
                raise InstantiateArgsError(
                    'Please provide instantiation arguments either in "args" or in the jenesis.toml configuration for this contract and profile'
                )
        validate(args, self._contract.schema[INSTANTIATE_MSG])
        address = super().instantiate(code_id, args, sender, label=label, gas_limit=gas_limit,
                                      admin_address=admin_address, funds=funds)

        if self._observer is not None:
            self._observer.on_contract_address_update(address)

        return address

    def deploy(
        self,
        args: Any,
        sender: Wallet,
        label: Optional[str] = None,
        store_gas_limit: Optional[int] = None,
        instantiate_gas_limit: Optional[int] = None,
        admin_address: Optional[Address] = None,
        funds: Optional[str] = None,
    ) -> Address:
        code_id = self.store(sender, gas_limit=store_gas_limit)
        address = self.instantiate(
            code_id,
            args,
            sender,
            label=label,
            gas_limit=instantiate_gas_limit,
            admin_address=admin_address,
            funds=funds
        )
        return address

    def execute(
        self,
        args: Any,
        sender: Wallet,
        gas_limit: Optional[int] = None,
        funds: Optional[str] = None,
    ) -> SubmittedTx:
        validate(args, self._contract.schema[EXECUTE_MSG])
        return super().execute(args, sender, gas_limit, funds)

    def query(self, args: Any) -> Any:
        validate(args, self._contract.schema[QUERY_MSG])
        return super().query(args)

    def _find_contract_id_by_digest_with_hint(self, code_id_hint: int) -> Optional[int]:

        # try and lookup the specified code id
        try:
            req = QueryCodeRequest(code_id=code_id_hint)
            resp = self._client.wasm.Code(req)

            if resp.code_info.data_hash == self.digest:
                return int(resp.code_info.code_id)
        except (grpc.RpcError, RuntimeError) as ex:
            not_found = False

            # pylint: disable=no-member
            if hasattr(ex, 'details') and 'not found' in ex.details() or 'not found' in str(ex):
                not_found = True

            if not not_found:
                raise

        return self._find_contract_id_by_digest(self._digest)

    def __repr__(self):
        return str(self._address)



import argparse
import os

from jenesis.config import Config
from jenesis.contracts.detect import detect_contracts


def add_contract_command(parser):

    add_contract_cmd = parser.add_parser("contract")
    add_contract_cmd.add_argument("template", help="The name of the template to use")
    add_contract_cmd.add_argument("name", help="The name of contract to create")
    add_contract_cmd.add_argument(
        "-b", "--branch", help="The name of the branch that should be used"
    )
    add_contract_cmd.set_defaults(handler=run_add_contract)


def run_add_contract(args: argparse.Namespace):
    template = args.template
    name = args.name
    branch = args.branch

    project_root = os.path.abspath(os.getcwd())
    contract_root = os.path.join(project_root, "contracts", name)

    # check that we are actually running the command from the project root
    if not os.path.exists(os.path.join(project_root, "jenesis.toml")):
        # pylint: disable=all
        print("Please run command from project root")
        return False

    # check to see if the contract already exists
    if os.path.exists(contract_root):
        print(f'Contract "{name}" already exists')
        return False

    Config.add_contract(contract_root, template, name, branch)

    contracts = detect_contracts(project_root)

    selected_contract = ""
    for contract in contracts:
        if contract.name == name:
            selected_contract = contract
            continue

    if selected_contract == "":
        print("Contract not found in project")
        return


    cfg = Config.load(project_root)

    for profile in cfg.profiles.keys():
        network_name = cfg.profiles[profile].network.name
        Config.update_project(os.getcwd(), profile, network_name, selected_contract)

    return True



import argparse
import os

from .shell import load_config


def run(args: argparse.Namespace):
    project_path = os.getcwd()

    # check that we are actually running the command from the project root
    if not os.path.exists(os.path.join(project_path, "jenesis.toml")):
        # pylint: disable=all
        print("Please run command from project root")
        return

    shell_globals = load_config(args)
    with open(args.script_path, encoding="utf-8") as script:
        exec(script.read(), shell_globals)


def add_run_command(parser):
    run_cmd = parser.add_parser("run")
    run_cmd.add_argument(
        "-p", "--profile", default=None, help="The profile to use"
    )
    run_cmd.add_argument("script_path", help="The path to the script to run")
    run_cmd.set_defaults(handler=run)



from concurrent.futures import ThreadPoolExecutor, TimeoutError as ConcurrentTimeoutError
from typing import Optional, Tuple, List

from cosmpy.aerial.client import LedgerClient
from cosmpy.aerial.contract import LedgerContract
from cosmpy.aerial.wallet import LocalWallet, Wallet
from cosmpy.crypto.address import Address
from cosmpy.crypto.keypairs import PrivateKey

from jenesis.config import Config, Deployment, Profile
from jenesis.contracts import Contract
from jenesis.contracts.detect import detect_contracts
from jenesis.contracts.monkey import MonkeyContract
from jenesis.keyring import query_keychain_item, LocalInfo, query_keychain_items
from jenesis.network import run_local_node
from jenesis.tasks import Task, TaskStatus
from jenesis.tasks.monitor import run_tasks
import os


class DeployContractTask(Task):
    def __init__(self, project_path: str, cfg: Config, profile: Profile, contract: Contract, config: Deployment,
                 client: LedgerClient, wallet: Wallet):
        self._project_path = project_path
        self._cfg = cfg
        self._profile = profile
        self._contract = contract
        self._config = config
        self._client = client
        self._wallet = wallet

        # task status
        self._status = TaskStatus.IDLE
        self._status_text = ''

        # background worker
        self._executor = ThreadPoolExecutor(max_workers=1)

        # state machine state
        self._state = 'idle'
        self.ledger_contract = None  # type: Optional[LedgerContract]
        self.contract_address = None  # type: Optional[Address]
        self._future = None

    @property
    def name(self) -> str:
        return self._contract.name

    @property
    def status(self) -> TaskStatus:
        return self._status

    @property
    def status_text(self) -> str:
        return self._status_text

    def poll(self):
        # print('POLL', self._state, self._status, self._status_text, self.name)

        if self._state == 'idle':
            self._schedule_build_contract()
        elif self._state == 'wait-for-ledger-contract':
            self._wait_for_ledger_contract()
        elif self._state == 'schedule-deployment':
            self._schedule_deploy_contract()
        elif self._state == 'wait-for-deployment':
            self._wait_for_contract_deployment()
        elif self._state == 'failed':
            self._failed()
        elif self._state == 'complete':
            self._complete()
        else:
            assert False, "bad state"

    def _schedule_build_contract(self):
        assert self._future is None

        def action():
            return MonkeyContract(
                self._contract,
                self._client,
                code_id=self._config.code_id,
                address=self._config.address
            )

        self._future = self._executor.submit(action)
        self._state = 'wait-for-ledger-contract'
        self._status = TaskStatus.IN_PROGRESS
        self._status_text = '(1/2) Determining contract parameters...'

    def _wait_for_ledger_contract(self):
        try:
            self.ledger_contract = self._future.result(timeout=0.05)
            self._state = 'schedule-deployment'
            self._future = None
        except ConcurrentTimeoutError:
            pass

    def _schedule_deploy_contract(self):
        assert self._future is None
        assert self.ledger_contract is not None

        def action():
            return self.ledger_contract.deploy(
                args=self._config.init,
                sender=self._wallet,
                admin_address=self._wallet.address(),
                funds=self._config.init_funds,
            )

        self._future = self._executor.submit(action)
        self._state = 'wait-for-deployment'
        self._status = TaskStatus.IN_PROGRESS
        self._status_text = '(2/2) Deploying contract...'

    def _wait_for_contract_deployment(self):
        try:
            self.contract_address = self._future.result(timeout=0.05)
            self._state = 'complete'
            self._future = None
        except ConcurrentTimeoutError:
            pass

    def _complete(self):
        # update the configuration and save it to disk
        self._cfg.update_deployment(self._profile.name, self._contract.name,
                                    self.ledger_contract.digest.hex(),
                                    self.ledger_contract.code_id, self.contract_address)
        self._cfg.save(self._project_path)

        self._finished(True)

    def _failed(self):
        self._finished(False)

    def _finished(self, success: bool):
        self._executor.shutdown(wait=True)
        if success:
            self._status = TaskStatus.COMPLETE
        else:
            self._status = TaskStatus.FAILED
        self._status_text = ''


def deploy_contracts(cfg: Config, project_path: str, deployer_key: Optional[str], profile: Optional[str] = None):
    # pylint: disable=all
    if profile is None:
        profile = cfg.get_default_profile()

    selected_profile = cfg.profiles[profile]

    if selected_profile.network.is_local:
        run_local_node(selected_profile.network)

    contracts = detect_contracts(project_path)

    contracts_to_deploy = []  # type: List[Tuple[Contract, Deployment]]
    profile_contracts = selected_profile.contracts
    profile_contract_names = list(profile_contracts.keys())

    # determine what tasks to do
    for contract in contracts:
        if contract.name in profile_contract_names:
            profile_contract = selected_profile.deployments.get(contract.name)
        else:
            continue
        assert profile_contract is not None

        # ensure that contract has been compiled first
        if not os.path.isfile(contract.binary_path):
            print(f"No contract binary found for {contract.name}. Please run 'jenesis compile' first.")
            continue

        if deployer_key is not None:
            profile_contract.deployer_key = deployer_key
            Config.update_key(os.getcwd(), profile, contract, deployer_key)

        # simple case the contract is already deployed and we can just use the information directly from the lockfile
        if profile_contract.is_configuration_out_of_date():
            contracts_to_deploy.append((contract, profile_contract))
            continue

        digest = contract.digest()
        if digest is None:
            continue  # we can't process any contracts where we don't have

        assert digest is not None

        # if the digest of the contract has changed then we need to add it to the list of contracts to deploy
        if profile_contract.digest != digest:
            contracts_to_deploy.append((contract, profile_contract))
            continue

    # exit if there is nothing to do
    if len(contracts_to_deploy) == 0:
        print('Nothing to deploy')
        return

    client = LedgerClient(selected_profile.network)

    # load all the keys required for this operation
    keys = {}
    available_key_names = set(query_keychain_items())
    all_keys = set(settings.deployer_key for _, settings in contracts_to_deploy)

    for key_name in all_keys:
        if key_name not in available_key_names:
            print(f'Unknown deployment key {key_name}')
            return

        info = query_keychain_item(key_name)
        if not isinstance(info, LocalInfo):
            print(f'Unable to lookup local key {key_name}')
            return

        keys[key_name] = PrivateKey(info.private_key)

    for contract, _ in contracts_to_deploy:
        # reset this contracts metadata
        contract_settings = selected_profile.deployments[contract.name]
        contract_settings.address = None  # clear the old address

        # lookup the wallet key
        wallet = LocalWallet(keys[contract_settings.deployer_key])

        # create the deployment task
        task = DeployContractTask(
            project_path,
            cfg,
            selected_profile,
            contract,
            contract_settings,
            client,
            wallet,
        )

        # run the deployment task
        run_tasks([task])

    return



import os.path
from dataclasses import dataclass
from typing import Optional

from cosmpy.aerial.contract import _compute_digest


@dataclass
class Contract:
    name: str
    source_path: str
    binary_path: str
    schema: dict

    def digest(self) -> Optional[str]:
        if not os.path.isfile(self.binary_path):
            return None

        return _compute_digest(self.binary_path).hex()

    def execute_msgs(self) -> dict:
        return self._extract_msgs('execute_msg')

    def query_msgs(self) -> dict:
        return self._extract_msgs('query_msg')

    def init_args(self) -> dict:
        return self.schema.get('instantiate_msg', {}).get('properties',{}).keys()

    def _extract_msgs(self, msg_type: str) -> dict:
        msgs = {}
        if 'oneOf' in self.schema[msg_type]:
            schemas = self.schema[msg_type]['oneOf']
        else:
            schemas = self.schema[msg_type]['anyOf']
        for schema in schemas:
            msg = schema['required'][0]
            if 'required' in schema['properties'][msg]:
                args = schema['properties'][msg]['required']
            else:
                args = []
            msgs[msg] = args
        return msgs

    def __repr__(self):
        return self.name



import argparse
import os

from cosmpy.aerial.client import LedgerClient
from jenesis.config import Config
from jenesis.contracts.detect import detect_contracts
from jenesis.contracts.monkey import MonkeyContract
from jenesis.network import Network
import toml

def run(args: argparse.Namespace):

    project_path = os.getcwd()

    # check that we are actually running the command from the project root
    if not os.path.exists(os.path.join(project_path, "jenesis.toml")):
        # pylint: disable=all
        print("Please run command from project root")
        return 1

    cfg = Config.load(project_path)

    if args.profile is not None and args.profile not in cfg.profiles:
        print(f'Invalid profile name. Expected one of {",".join(cfg.profiles.keys())}')
        return 1

    profile_name = args.profile or cfg.get_default_profile()

    selected_profile = cfg.profiles[profile_name]

    contracts = detect_contracts(project_path)

    selected_contract = ""
    for contract in contracts:
        if contract.name == args.contract:
            selected_contract = contract
            continue

    if selected_contract == "":
        print('Contract not found in project')
        return 1

    client = LedgerClient(selected_profile.network)

    contract = MonkeyContract(selected_contract, client, args.address)
    code_id = contract.code_id
    digest = contract.digest.hex()

    data = toml.load("jenesis.toml") 

    network = Network(**data["profile"][profile_name]["network"])

    Config.update_project(project_path, profile_name, network.name, selected_contract)

    cfg.update_deployment(
        selected_profile.name, args.contract, digest, code_id, args.address
    )
    cfg.save(project_path)
    return 0


def add_attach_command(parser):
    attach_cmd = parser.add_parser("attach")
    attach_cmd.add_argument("contract", help="Contract name")
    attach_cmd.add_argument("address", help="Contract address")
    attach_cmd.add_argument(
        "-p",
        "--profile",
        default=None,
        help="Profile where the contract will be attached",
    )
    attach_cmd.set_defaults(handler=run)


import os
import tempfile
import time
from typing import Optional, List

from docker import from_env
from docker.errors import DockerException
from docker.types import Mount

from cosmpy.aerial.config import NetworkConfig
from cosmpy.aerial.client import LedgerClient

DEFAULT_DOCKER_IMAGE_TAG = "fetchai/fetchd:0.10.5"
DEFAULT_VALIDATOR_KEY_NAME = "validator"
DEFAULT_MNEMONIC = "gap bomb bulk border original scare assault pelican resemble found laptop skin gesture height inflict clinic reject giggle hurdle bubble soldier hurt moon hint"
DEFAULT_PASSWORD = "12345678"
DEFAULT_MONIKER = "test-node"
DEFAULT_CHAIN_ID = "testing"
DEFAULT_GENESIS_ACCOUNT = "fetch1gns5lphdk5ew5lnre7ulzv8s8k9dr9eyqvgj0w"
DEFAULT_DENOMINATION = "atestfet"
DEFAULT_CLI_BINARY = "fetchd"

TMP_DIR = tempfile.TemporaryDirectory() # pylint: disable=consider-using-with

class Network(NetworkConfig):

    def __init__(
        self,
        name: str = "",
        chain_id: str = "",
        fee_minimum_gas_price: int = 0,
        fee_denomination: str = "",
        staking_denomination: str = "",
        url: str = "",
        faucet_url: Optional[str] = None,
        is_local: Optional[bool] = False,
        cli_binary: Optional[str] = None,
        validator_key_name: Optional[str] = None,
        mnemonic: Optional[str] = None,
        password: Optional[str] = None,
        moniker: Optional[str] = None,
        genesis_accounts: Optional[List[str]] = None,
    ):
        super().__init__(
            chain_id,
            fee_minimum_gas_price,
            fee_denomination,
            staking_denomination,
            url,
            faucet_url=faucet_url,
        )
        self.name = name
        self.is_local = is_local
        if is_local:
            self.cli_binary = cli_binary or DEFAULT_CLI_BINARY
            self.validator_key_name = validator_key_name or DEFAULT_VALIDATOR_KEY_NAME
            self.mnemonic = mnemonic or DEFAULT_MNEMONIC
            self.password = password or DEFAULT_PASSWORD
            self.moniker = moniker or DEFAULT_MONIKER
            self.genesis_accounts = genesis_accounts or [DEFAULT_GENESIS_ACCOUNT]

    @classmethod
    def from_cosmpy_config(cls, name: str, net_config: NetworkConfig, is_local: bool = False):
        return Network(
            name,
            net_config.chain_id,
            net_config.fee_minimum_gas_price,
            net_config.fee_denomination,
            net_config.staking_denomination,
            net_config.url,
            faucet_url=net_config.faucet_url,
            is_local=is_local,
        )


class LedgerNodeDockerContainer:

    PORTS = {9090: 9090, 1317: 1317, 26657: 26657}

    def __init__(
        self,
        network: Network,
        tag: str = DEFAULT_DOCKER_IMAGE_TAG,
    ):
        """
        Initialize the Fetch ledger Docker image.

        :param client: the Docker client.
        :param addr: the address.
        :param port: the port.
        :param config: optional configuration to command line.
        """
        self._client = from_env()
        self._image_tag = tag
        self.network: Network = network

    @property
    def tag(self) -> str:
        """Get the image tag."""
        return self._image_tag

    def _make_entrypoint_file(self, tmpdirname) -> None:
        """Make a temporary entrypoint file to setup and run the test ledger node"""
        run_node_lines = [
            "#!/usr/bin/env bash",
            # variables
            f'export VALIDATOR_KEY_NAME={self.network.validator_key_name}',
            f'export VALIDATOR_MNEMONIC="{self.network.mnemonic}"',
            f'export PASSWORD="{self.network.password}"',
            f'export CHAIN_ID={self.network.chain_id}',
            f'export MONIKER={self.network.moniker}',
            f'export DENOM={self.network.fee_denomination}',
            # Add key
            f'( echo "$VALIDATOR_MNEMONIC"; echo "$PASSWORD"; echo "$PASSWORD"; ) |{self.network.cli_binary} keys add $VALIDATOR_KEY_NAME --recover',
            # Configure node
            f"{self.network.cli_binary} init --chain-id=$CHAIN_ID $MONIKER",
        ]
        for acc in self.network.genesis_accounts:
            run_node_lines.append(
                f'echo "$PASSWORD" |{self.network.cli_binary} add-genesis-account {acc} 100000000000000000000000$DENOM',
            )
        run_node_lines.extend([
            f'echo "$PASSWORD" |{self.network.cli_binary} add-genesis-account $({self.network.cli_binary} keys show $VALIDATOR_KEY_NAME -a) 100000000000000000000000$DENOM',
            f'echo "$PASSWORD" |{self.network.cli_binary} gentx $VALIDATOR_KEY_NAME 10000000000000000000000$DENOM --chain-id $CHAIN_ID',
            f"{self.network.cli_binary} collect-gentxs",
            # Enable rest-api
            f'sed -i "s/stake/atestfet/" ~/.{self.network.cli_binary}/config/genesis.json',
            f'sed -i "s/enable = false/enable = true/" ~/.{self.network.cli_binary}/config/app.toml',
            f'sed -i "s/swagger = false/swagger = true/" ~/.{self.network.cli_binary}/config/app.toml',
            f"{self.network.cli_binary} start --rpc.laddr tcp://0.0.0.0:26657",
        ])
        entrypoint_file = os.path.join(tmpdirname, "run-node.sh")
        with open(entrypoint_file, "w", encoding="utf-8") as file:
            file.writelines(line + "\n" for line in run_node_lines)
        os.chmod(entrypoint_file, 300)

    def run(self):
        self._make_entrypoint_file(TMP_DIR.name)
        mount_path = "/mnt"
        volumes = {TMP_DIR.name: {"bind": mount_path, "mode": "rw"}}
        entrypoint = os.path.join(mount_path, "run-node.sh")
        container = self._client.containers.run(
            self.tag,
            detach=True,
            volumes=volumes,
            entrypoint=str(entrypoint),
            ports=self.PORTS,
        )
        return container

    @classmethod
    def is_ready(cls) -> bool:
        try:
            net_config = fetchai_localnode_config()
            client = LedgerClient(net_config)
            validators = client.query_validators()
            assert len(validators) > 0
            return True
        except Exception:
            return False

    @classmethod
    def wait_until_ready(cls, max_attempts: int = 15, sleep_rate: float = 1.0) -> bool:
        for _ in range(max_attempts):
            if cls.is_ready():
                return True
            time.sleep(sleep_rate)
        return False


def run_local_node(network: Network):
    try:
        local_node = LedgerNodeDockerContainer(network)
        if not local_node.is_ready():
            print("Starting local node...")
            container = local_node.run()
            if not container.status == "created":
                raise RuntimeError('Failed to create local node.')
            if not local_node.wait_until_ready():
                raise RuntimeError('Failed to start local node.')
            print("Stating local node...complete")
        else:
            print("Detected local node already running.")
    except DockerException:
        print("Failed to start local node: looks like your docker setup isn't right, please visit https://jenesis.fetch.ai/ for more information")


def fetchai_testnet_config() -> Network:
    return Network.from_cosmpy_config("fetchai-testnet", NetworkConfig.fetchai_stable_testnet())


def fetchai_localnode_config() -> Network:
    return Network.from_cosmpy_config(
        "fetchai-localnode",
        NetworkConfig(
            chain_id="localnode",
            url="grpc+http://127.0.0.1:9090/",
            fee_minimum_gas_price=5000000000,
            fee_denomination="atestfet",
            staking_denomination="atestfet",
        ),
        is_local=True,
    )



import fnmatch
import os
from typing import List, Optional, Any

from docker import from_env
from docker.errors import DockerException
from docker.types import Mount

from jenesis.contracts import Contract
from jenesis.tasks import Task, TaskStatus
from jenesis.tasks.monitor import run_tasks
from jenesis.config import Config

class ContractBuildTask(Task):

    BUILD_CONTAINER = 'cosmwasm/rust-optimizer:0.12.5'

    def __init__(self, contract: Contract):
        self.contract = contract
        self._container = None
        self._status = TaskStatus.IDLE
        self._status_text = ''

    @property
    def name(self) -> str:
        return self.contract.name

    @property
    def status_text(self) -> str:
        return self._status_text

    @property
    def status(self) -> TaskStatus:
        return self._status

    def poll(self):
        # no further processing required if it has either completed or failed
        if self.is_done:
            return

        # if the contract is not out of date then no build is required
        if self._container is None and not self._is_out_of_date():
            self._status = TaskStatus.COMPLETE
            self._status_text = ''
            return

        # if we get this far we either need to schedule a docker build of the contract or we need to monitor
        # the progress of a docker build
        if self._container is None:
            cfg = Config.load(os.getcwd())
            try:
                self._container = self._schedule_build_container(self.contract)
                self._status = TaskStatus.IN_PROGRESS
                self._status_text = 'Building...'
                for profile in cfg.profiles.keys():
                    # pylint: disable=all
                    network_name = cfg.profiles[profile].network.name
                    Config.update_project(os.getcwd(), profile, network_name, self.contract)
            except DockerException:
                print("Error: looks like your docker setup isn't right, please visit https://jenesis.fetch.ai/ for more information")
                self._container = None
                self._status = TaskStatus.FAILED
                self._status_text = ''

            # exit if we do not have docker or some such installed
            if self._container is None:
                return

        assert self._container is not None

        # check on the progress of the container
        self._container.reload()

        if self._container.status == 'exited':
            exit_code = int(self._container.attrs['State']['ExitCode'])
            if exit_code == 0:
                self._status = TaskStatus.COMPLETE
                self._status_text = ''

                # clean up the container if it was successful, otherwise keep if for the logs
                self._container.remove()

            else:
                self._status = TaskStatus.FAILED
                self._status_text = ''

    def _is_out_of_date(self) -> bool:
        # determine the timestamp of the compiled contract
        if os.path.isfile(self.contract.binary_path):
            compiled_contract_timestamp = os.path.getmtime(self.contract.binary_path)
        else:
            compiled_contract_timestamp = 0

        # determine the timestamp of the contract source
        contract_source_timestamp = self._get_contract_modified_timestamp(self.contract.source_path)

        return contract_source_timestamp > compiled_contract_timestamp

    @classmethod
    def _schedule_build_container(cls, contract: Contract):
        mounts = [
            Mount('/code/target', f'contract_{contract.name}_cache'),
            Mount('/usr/local/cargo/registry', 'registry_cache'),
            Mount('/code', os.path.abspath(contract.source_path), type='bind'),
        ]

        # get the docker client
        client = from_env()

        # start the container
        return client.containers.run(cls.BUILD_CONTAINER, mounts=mounts, detach=True)

    @staticmethod
    def _get_src_files(path: str):
        for root, _, files in os.walk(path):
            for filename in fnmatch.filter(files, '*.rs'):
                yield os.path.join(root, filename)

    @classmethod
    def _get_contract_modified_timestamp(cls, path: str) -> float:
        src_path = os.path.join(path, 'src')
        return max(
            map(
                os.path.getmtime,
                cls._get_src_files(src_path)
            ),
            default=0
        )


def _chunks(values: List[Any], batch_size: Optional[int]):
    if batch_size is None:
        yield values
    else:
        for i in range(0, len(values), batch_size):
            yield values[i:i + batch_size]


def build_contracts(contracts: List[Contract], batch_size: Optional[int] = None):
    """
    Will attempt to build all the specified contracts (provided they are out of date)

    :param contracts: The list of contracts to build
    :param batch_size: The max number of builds to do in parallel. If None then will attempt to all in parallel
    :return:
    """
    # create all the tasks to be done
    tasks = list(
        map(
            ContractBuildTask,
            contracts,
        )
    )

    # run the tasks (in batches if configured)
    for batch in _chunks(tasks, batch_size=batch_size):
        run_tasks(batch)


class WorkspaceBuildTask(Task):

    BUILD_CONTAINER = 'cosmwasm/workspace-optimizer:0.12.5'

    def __init__(self, path: str, contracts: List[Contract]):
        self._path = path
        self._contracts = contracts
        self._container = None
        self._status = TaskStatus.IDLE
        self._status_text = ''

    @property
    def name(self) -> str:
        return os.path.basename(self._path)

    @property
    def path(self) -> str:
        return self._path

    @property
    def status_text(self) -> str:
        return self._status_text

    @property
    def status(self) -> TaskStatus:
        return self._status

    def poll(self):
        # no further processing required if it has either completed or failed
        if self.is_done:
            return

        # if the contract is not out of date then no build is required
        if self._container is None and not self._is_out_of_date():
            self._status = TaskStatus.COMPLETE
            self._status_text = ''
            return

        # if we get this far we either need to schedule a docker build of the workspace or we need to monitor
        # the progress of a docker build
        if self._container is None:
            cfg = Config.load(os.getcwd())
            try:
                self._container = self._schedule_build_container(self._path)
                self._status = TaskStatus.IN_PROGRESS
                self._status_text = 'Building...'

                for profile in cfg.profiles.keys():
                    # pylint: disable=all
                    network_name = cfg.profiles[profile].network.name
                    for contract in self._contracts:
                        Config.update_project(os.getcwd(), profile, network_name, contract)
                        
            except DockerException:
                self._container = None
                self._status = TaskStatus.FAILED
                self._status_text = ''

            # exit if we do not have docker or some such installed
            if self._container is None:
                return

        assert self._container is not None

        # check on the progress of the container
        self._container.reload()

        if self._container.status == 'exited':
            exit_code = int(self._container.attrs['State']['ExitCode'])
            if exit_code == 0:
                self._status = TaskStatus.COMPLETE
                self._status_text = ''

                # clean up the container if it was successful, otherwise keep if for the logs
                self._container.remove()

            else:
                self._status = TaskStatus.FAILED
                self._status_text = ''

    def _is_out_of_date(self) -> bool:
        # determine the most recent timestamp of the compiled workspace files
        workspace_build_timestamp = self._get_workspace_build_timestamp(self._path)

        # determine the timestamp of the contract source
        contract_source_timestamp = self._get_contract_modified_timestamp(self._contracts)

        return contract_source_timestamp > workspace_build_timestamp

    @classmethod
    def _schedule_build_container(cls, path: str):
        mounts = [
            Mount('/code/target', f'workspace_{os.path.basename(path)}_cache'),
            Mount('/usr/local/cargo/registry', 'registry_cache'),
            Mount('/code', os.path.abspath(path), type='bind'),
        ]

        # get the docker client
        client = from_env()

        # start the container
        return client.containers.run(cls.BUILD_CONTAINER, mounts=mounts, detach=True)

    @staticmethod
    def _get_src_files(contracts: List[Contract]):
        for contract in contracts:
            src_path = os.path.join(contract.source_path, 'src')
            for root, _, files in os.walk(src_path):
                for filename in fnmatch.filter(files, '*.rs'):
                    yield os.path.join(root, filename)

    @staticmethod
    def _get_build_files(path: str):
        for root, _, files in os.walk(path):
            for filename in fnmatch.filter(files, '*.wasm'):
                yield os.path.join(root, filename)

    @classmethod
    def _get_contract_modified_timestamp(cls, contracts: List[Contract]) -> float:
        return max(
            map(
                os.path.getmtime,
                cls._get_src_files(contracts)
            ),
            default=0
        )

    @classmethod
    def _get_workspace_build_timestamp(cls, path: str) -> float:
        build_path = os.path.join(path, 'artifacts')
        return max(
            map(
                os.path.getmtime,
                cls._get_build_files(build_path)
            ),
            default=0
        )


def build_workspace(path: str, contracts: List[Contract]):
    """
    Will attempt to build the cargo workspace including all contracts

    :return:
    """
    # create all the tasks to be done
    tasks = [WorkspaceBuildTask(path, contracts)]

    # run the tasks
    run_tasks(tasks)



import os
from typing import List, Optional

import json
import toml

from jenesis.contracts import Contract


def is_workspace(path: str) -> bool:
    contracts_folder = os.path.join(path, 'contracts')
    if not os.path.isdir(contracts_folder):
        return False

    expected_files = (
        'Cargo.toml',
        'Cargo.lock',
    )
    for expected_file in expected_files:
        if not os.path.isfile(os.path.join(path, expected_file)):
            return False

    return True


def detect_contracts(path: str) -> Optional[List[Contract]]:
    contracts_folder = os.path.join(path, 'contracts')
    if not os.path.isdir(contracts_folder):
        return None

    def is_contract(folder_name: str) -> bool:
        contract_path = os.path.join(contracts_folder, folder_name)
        if not os.path.isdir(contract_path):
            return False

        expected_files = (
            'Cargo.toml',
        )
        for expected_file in expected_files:
            if not os.path.isfile(os.path.join(contract_path, expected_file)):
                return False

        return True

    def parse_contract(name: str) -> Contract:
        cargo_file_path = os.path.join(contracts_folder, name, 'Cargo.toml')
        with open(cargo_file_path, 'r', encoding="utf-8") as cargo_file:
            cargo_contents = toml.load(cargo_file)

        # extract the contract name
        contract_name = cargo_contents['package']['name']

        source_path = os.path.abspath(os.path.join(contracts_folder, name))
        if is_workspace(path):
            artifacts_path = os.path.join(path, 'artifacts')
        else:
            artifacts_path = os.path.join(contracts_folder, name, 'artifacts')
        binary_path = os.path.abspath(os.path.join(artifacts_path, f'{contract_name}.wasm'))
        schema = load_contract_schema(source_path)

        return Contract(
            name=name,
            source_path=source_path,
            binary_path=binary_path,
            schema=schema,
        )

    contracts = map(
        parse_contract,
        filter(
            is_contract,
            os.listdir(contracts_folder)
        )
    )

    return list(contracts)


def load_contract_schema(source_path: str) -> dict:
    schema_folder = os.path.join(source_path, 'schema')
    if not os.path.isdir(schema_folder):
        return None

    schema = {}
    for filename in os.listdir(schema_folder):
        if filename.endswith('.json'):
            msg_name = os.path.splitext(os.path.basename(filename))[0]
            full_path = os.path.join(schema_folder, filename)
            with open(full_path, 'r', encoding="utf-8") as msg_schema_file:
                msg_schema = json.load(msg_schema_file)
            schema[msg_name] = msg_schema
    return schema



import argparse
import os

from blessings import Terminal
from jenesis.contracts.build import build_contracts, build_workspace
from jenesis.contracts.detect import detect_contracts, is_workspace


def run(args: argparse.Namespace):

    project_path = os.getcwd()

    # check that we are actually running the command from the project root
    if not os.path.exists(os.path.join(project_path, "jenesis.toml")):
        print("Please run command from project root")
        return 1

    term = Terminal()

    contracts = detect_contracts(project_path)
    if contracts is None or len(contracts) == 0:
        print(term.red("Unable to detect any contracts"))
        return 1

    if is_workspace(project_path):
        print(term.green("Detected cargo workspace"))
        build_workspace(project_path, contracts)
        return 0

    # build the contracts
    build_contracts(contracts, batch_size=args.batch_size)
    return 0


def add_compile_command(parser):
    compile_cmd = parser.add_parser("compile")
    compile_cmd.add_argument(
        "-p",
        dest="batch_size",
        type=int,
        help="The limit of the number of tasks to do in parallel",
    )
    compile_cmd.set_defaults(handler=run)



import platform
from typing import List, Union

from jenesis.keyring.amino.codec import LocalInfo, OfflineInfo

try:
    from jenesis.keyring import macos
except Exception:
    pass


try:
    from jenesis.keyring import linux
except Exception:
    pass


def query_keychain_item(name: str) -> Union[LocalInfo, OfflineInfo]:
    system = platform.system()
    if system == "Darwin":
        return macos.query_keychain_item(name)

    if system == "Linux":
        return linux.query_keychain_item(name)

    raise RuntimeError(f"Platform {system} not currently supported")


def query_keychain_items() -> List[str]:
    system = platform.system()
    if system == "Darwin":
        return macos.query_keychain_items()

    if system == "Linux":
        return linux.query_keychain_items()

    raise RuntimeError(f"Platform {system} not currently supported")
